//
// wavwriter.h
// Copyright Phase Five Systems
//
#include <stdint.h>
#include <cstdio>

#include "jump/protocol/base/rdclientbase.h"
#include "jump/protocol/rtc/audiocapture.h"

namespace jump
{

class WaveWriter
{
public:
    struct WAV_HEADER
    {
        /* RIFF Chunk Descriptor */
        uint8_t RIFF[4] = {'R', 'I', 'F', 'F'}; // RIFF Header Magic header
        uint32_t ChunkSize = 0;                 // RIFF Chunk Size
        uint8_t WAVE[4] = {'W', 'A', 'V', 'E'}; // WAVE Header
        /* "fmt" sub-chunk */
        uint8_t fmt[4] = {'f', 'm', 't', ' '}; // FMT header
        uint32_t Subchunk1Size = 16;           // Size of the fmt chunk
        uint16_t AudioFormat = 1; // Audio format 1=PCM,6=mulaw,7=alaw, 257=IBM
                                  // Mu-Law, 258=IBM A-Law, 259=ADPCM
        uint16_t NumOfChan = 0;     // Number of channels 1=Mono 2=Sterio
        uint32_t SamplesPerSec = 0; // Sampling Frequency in Hz
        uint32_t bytesPerSec = 0;   // bytes per second
        uint16_t blockAlign = 0;    // 2=16-bit mono, 4=16-bit stereo
        uint16_t bitsPerSample = 0; // Number of bits per sample
        /* "data" sub-chunk */
        uint8_t Subchunk2ID[4] = {'d', 'a', 't', 'a'}; // "data"  string
        uint32_t Subchunk2Size = 0;                    // Sampled data length
    };

    static void WriteHeader(FILE *s,
                            uint32_t numChans,
                            uint32_t sampleRateHZ,
                            uint32_t bitsPerSample,
                            uint32_t dataSize)
    {
        WAV_HEADER header;
        header.ChunkSize = sizeof(header) + dataSize - 8;
        header.NumOfChan = numChans;
        header.SamplesPerSec = sampleRateHZ;
        header.bytesPerSec = numChans * sampleRateHZ * (bitsPerSample / 8);
        header.blockAlign = numChans * (bitsPerSample / 8);
        header.bitsPerSample = bitsPerSample;
        header.Subchunk2Size = dataSize;
        fwrite(reinterpret_cast<const void *>(&header), sizeof(header), 1, s);
    }

    static void Append(FILE *f,
                       uint32_t numChans,
                       uint32_t sampleRateHZ,
                       uint32_t bitsPerSample,
                       const uint8_t *data,
                       uint32_t dataSize)
    {
        fseek(f, 0, SEEK_END);
        uint64_t fileSize = ftell(f);
        if (fileSize == 0)
        {
            // Write an empty header
            WriteHeader(f, numChans, sampleRateHZ, bitsPerSample, 0);
            fileSize += sizeof(WAV_HEADER);
        }

        // Write the samples to the end of the file
        fwrite(reinterpret_cast<const void *>(data), dataSize, 1, f);

        fileSize += dataSize;

        // Overwrite header with new data
        fseek(f, 0, SEEK_SET);
        WriteHeader(f,
                    numChans,
                    sampleRateHZ,
                    bitsPerSample,
                    fileSize - sizeof(WAV_HEADER));
    }

    static void Append(const std::string &path,
                       uint32_t numChans,
                       uint32_t sampleRateHZ,
                       uint32_t bitsPerSample,
                       const uint8_t *data,
                       uint32_t dataSize)
    {
        FILE *f = fopen(path.c_str(), "rb+");
        if (f == nullptr)
        {
            f = fopen(path.c_str(), "wb+");
        }

        if (f)
        {
            Append(f, numChans, sampleRateHZ, bitsPerSample, data, dataSize);
            fclose(f);
        }
    }

    static void Append(const std::string &path,
                       const LocalAudioCaptureInterface::PCMAudioFormat &format,
                       const uint8_t *samples,
                       uint32_t sampleSizeBytes)
    {
        Append(path,
               format.NumChannels,
               format.SampleRateHZ,
               format.BitsPerSample,
               samples,
               sampleSizeBytes);
    }

    static void Append(const std::string &path,
                       const RDClientAudioStreamDescription &desc,
                       const uint8_t *samples,
                       uint32_t sampleSizeBytes)
    {
        Append(path,
               desc.numChannels,
               desc.sampleRateHZ,
               desc.bitsPerSample,
               samples,
               sampleSizeBytes);
    }
};

} // namespace jump
